﻿using System;
using System.Threading;
using System.Collections.Generic;
using System.Reflection;
using System.IO;
using System.Diagnostics;

using Reflector;
using Reflector.CodeModel;

namespace CSLint
{
    internal class CSLintAddin : IPackage
    {
        IWindowManager windowManager;
        IAssemblyManager assemblyManager;
        ICommandBarManager commandBarManager;
        ICommandBarSeparator separator;
        ICommandBarButton button;
        ITranslatorManager translatorManager;
        ITranslator translator;

        public void Load(IServiceProvider serviceProvider)
        {
            this.windowManager = (IWindowManager)serviceProvider.GetService(typeof(IWindowManager));
            //this.windowManager.ShowMessage("Loading CSLint!");
            this.assemblyManager = (IAssemblyManager)serviceProvider.GetService(typeof(IAssemblyManager));
            this.commandBarManager = (ICommandBarManager)serviceProvider.GetService(typeof(ICommandBarManager));
            this.translatorManager = (ITranslatorManager)serviceProvider.GetService(typeof(ITranslatorManager));
            this.translator = this.translatorManager.CreateDisassembler(null, null);
            this.separator = this.commandBarManager.CommandBars["Tools"].Items.AddSeparator();
            this.button = this.commandBarManager.CommandBars["Tools"].Items.AddButton("&Detect Deadlocks", new EventHandler(this.CSLintListButton_Click));
        }


        public void Unload()
        {
            //this.windowManager.ShowMessage("Unloading CSLint!");
            this.commandBarManager.CommandBars["Tools"].Items.Remove(this.button);
            this.commandBarManager.CommandBars["Tools"].Items.Remove(this.separator);
        }



        private void CSLintListButton_Click(object sender, EventArgs e)
        {
            IAssemblyCollection assemblyCollection = this.assemblyManager.Assemblies;
            List<IAssembly> assemblies = new List<IAssembly>(assemblyCollection.Count);
            foreach (IAssembly assembly in assemblyCollection)
            {
                assemblies.Add(assembly);
            }
			ITranslator translator = this.translatorManager.CreateDisassembler(null, null);
            foreach (IAssembly assembly in assemblies)
            {
                if (!assembly.Location.StartsWith("%SystemRoot%")
                    && !assembly.Location.StartsWith("%ProgramFiles%"))
                {
                    foreach (IModule module in assembly.Modules)
                    {
                        foreach (ITypeDeclaration type in module.Types)
                        {
                            processType(type);
                        }
                    }
                }
            }
            build();
            verify();
            Reset();
        }

        internal void build()
        {
            foreach (ClassDesc desc in classes.Values)
            {
                desc.addOverriddenMethodInvocations();
            }
            for (MethodDesc entry = entryPoints; entry != null; entry = entry.nextEntryPoint)
            {
                entry.buildConcurrentClosure();
            }
        }

        internal void verify()
        {
            String msg = GraphVertex.verify();
            message(msg);
            StreamWriter writer = new StreamWriter("deadlock.lst");
            writer.Write(msg);
            writer.Close();
        }

        internal MethodDesc getMethodDesc(IMethodReference mr)
        {
            if (mr != null)
            {
                IMethodDeclaration md = mr.Resolve();
                if (md != null)
                {
                    ClassDesc cls = getClassDesc(md.DeclaringType);
                    if (cls != null)
                    {
                        return cls.getMethod(md);
                    }
                }
            }
            return null;
        }

        internal ClassDesc getClassDesc(IType type)
        {
            ITypeReference typeRef = type as ITypeReference;
            if (typeRef != null)
            {
                ITypeDeclaration typeDecl = typeRef.Resolve();
                if (typeDecl != null)
                {
                    ClassDesc cls;
                    if (!classes.TryGetValue(typeDecl, out cls))
                    {
                        cls = new ClassDesc(typeDecl);
                        classes[typeDecl] = cls;
                    }
                    return cls;
                }
            }
            return null;
        }

        void processType(ITypeDeclaration type)
        {
            foreach (IMethodDeclaration mth in type.Methods)
            {
                processMethod(mth);
            }
            /*
            foreach (IPropertyDeclaration prop in type.Properties)
            {
                processMethod(prop.GetMethod);
                processMethod(prop.SetMethod);
            }
            */
            foreach (ITypeDeclaration nestedType in type.NestedTypes)
            {
                processType(nestedType);
            }
        }

        void processStatement(IStatement stmt)
        {
            if (stmt is IBlockStatement)
            {
                foreach (IStatement s in ((IBlockStatement)stmt).Statements)
                {
                    processStatement(s);
                }
            }
            else if (stmt is IWhileStatement)
            {
                processExpression(((IWhileStatement)stmt).Condition);
                processStatement(((IWhileStatement)stmt).Body);
            }
            else if (stmt is IForEachStatement)
            {
                processExpression(((IForEachStatement)stmt).Expression);
                processStatement(((IForEachStatement)stmt).Body);
            }
            else if (stmt is IForStatement)
            {
                processExpression(((IForStatement)stmt).Condition);
                processStatement(((IForStatement)stmt).Increment);
                processStatement(((IForStatement)stmt).Initializer);
                processStatement(((IForStatement)stmt).Body);
            }
            else if (stmt is ITryCatchFinallyStatement)
            {
                processStatement(((ITryCatchFinallyStatement)stmt).Try);
                processStatement(((ITryCatchFinallyStatement)stmt).Fault);
                processStatement(((ITryCatchFinallyStatement)stmt).Finally);
                foreach (ICatchClause clause in ((ITryCatchFinallyStatement)stmt).CatchClauses)
                {
                    processExpression(clause.Condition);
                    processStatement(clause.Body);
                }
            }
            else if (stmt is IConditionStatement)
            {
                processExpression(((IConditionStatement)stmt).Condition);
                processStatement(((IConditionStatement)stmt).Then);
                processStatement(((IConditionStatement)stmt).Else);
            }
            else if (stmt is ISwitchStatement)
            {
                processExpression(((ISwitchStatement)stmt).Expression);
                foreach (ISwitchCase cases in ((ISwitchStatement)stmt).Cases)
                {
                    processStatement(cases.Body);
                }
            }
            else if (stmt is IUsingStatement)
            {
                processExpression(((IUsingStatement)stmt).Expression);
                processStatement(((IUsingStatement)stmt).Body);
            }
            else if (stmt is ILockStatement)
            {
                IExpression expr = ((ILockStatement)stmt).Expression;
                processExpression(expr);
                object syncObj = null;
                if (expr is IMethodInvokeExpression)
                {
                    IExpression call = ((IMethodInvokeExpression)expr).Method;
                    if (call is IMethodReferenceExpression)
                    {
                        syncObj = ((IMethodReferenceExpression)call).Method;
                    }
                }
                else if (expr is IPropertyReferenceExpression)
                {
                    syncObj = ((IPropertyReferenceExpression)expr).Property;
                }
                else if (expr is IFieldReferenceExpression)
                {
                    syncObj = ((IFieldReferenceExpression)expr).Field;
                }
                else if (expr is IThisReferenceExpression)
                {
                    syncObj = currMethod.cls.cls;
                }
                else if (expr is IVariableReferenceExpression)
                {
                    IVariableDeclaration v = ((IVariableReferenceExpression)expr).Variable.Resolve();
                    if (v != null)
                    {
                        syncObj = v.VariableType as ITypeReference;
                    }
                }
                else if (expr is ITypeOfExpression)
                {
                    syncObj = expr;
                }
                if (syncObj != null)
                {
                    GraphVertex monitor;
                    if (!syncObjects.TryGetValue(syncObj, out monitor))
                    {
                        monitor = new GraphVertex(syncObj);
                        syncObjects[syncObj] = monitor;
                    }
                    if (monitors.empty())
                    {
                        topLevelLocks.push(monitor);
                        lockStmts.Add((ILockStatement)stmt);
                    }
                    else
                    {
                        monitors.top().addEdge(monitor, currMethod, stmt, CallContext.monitorEnter);
                    }
                    monitors.push(monitor);
                    processStatement(((ILockStatement)stmt).Body);
                    monitors.pop();
                }
                else
                {
                    processStatement(((ILockStatement)stmt).Body);
                }
            }
            else if (stmt is ILabeledStatement)
            {
                processStatement(((ILabeledStatement)stmt).Statement);
            }
            else if (stmt is IDoStatement)
            {
                processStatement(((IDoStatement)stmt).Body);
                processExpression(((IDoStatement)stmt).Condition);
            }
            else if (stmt is IExpressionStatement)
            {
                processExpression(((IExpressionStatement)stmt).Expression);
            }
        }

        void processExpression(IExpression expr)
        {
            if (expr is IMethodInvokeExpression)
            {
                IMethodInvokeExpression call = (IMethodInvokeExpression)expr;
                IExpression target = call.Method;
                if (target is IMethodReferenceExpression)
                {
                    IMethodReferenceExpression methodRef = (IMethodReferenceExpression)target;
                    MethodDesc callee = getMethodDesc(methodRef.Method);
                    if (callee != null)
                    {
                        currMethod.addInvocation(
                            callee,
                            !(methodRef.Target is IBaseReferenceExpression),
                            expr,
                            monitors);
                    }
                }
                processExpression(call.Method);
                foreach (IExpression arg in call.Arguments)
                {
                    processExpression(arg);
                }
            }
            else if (expr is IAssignExpression)
            {
                processExpression(((IAssignExpression)expr).Target);
                processExpression(((IAssignExpression)expr).Expression);
            }
            else if (expr is IBinaryExpression)
            {
                processExpression(((IBinaryExpression)expr).Left);
                processExpression(((IBinaryExpression)expr).Right);
            }
            else if (expr is IUnaryExpression)
            {
                processExpression(((IUnaryExpression)expr).Expression);
            }
            else if (expr is ICastExpression)
            {
                processExpression(((ICastExpression)expr).Expression);
            }
            else if (expr is IArrayIndexerExpression)
            {
                processExpression(((IArrayIndexerExpression)expr).Target);
                foreach (IExpression index in ((IArrayIndexerExpression)expr).Indices)
                {
                    processExpression(index);
                }
            }
            else if (expr is IArrayCreateExpression)
            {
                processExpression(((IArrayCreateExpression)expr).Initializer);
                foreach (IExpression dim in ((IArrayCreateExpression)expr).Dimensions)
                {
                    processExpression(dim);
                }
            }
            else if (expr is ICanCastExpression)
            {
                processExpression(((ICanCastExpression)expr).Expression);
            }
            else if (expr is IConditionExpression)
            {
                processExpression(((IConditionExpression)expr).Condition);
                processExpression(((IConditionExpression)expr).Then);
                processExpression(((IConditionExpression)expr).Else);
            }
            else if (expr is IDelegateInvokeExpression)
            {
                processExpression(((IDelegateInvokeExpression)expr).Target);
                foreach (IExpression arg in ((IDelegateInvokeExpression)expr).Arguments)
                {
                    processExpression(arg);
                }
            }
            else if (expr is IEventReferenceExpression)
            {
                processExpression(((IEventReferenceExpression)expr).Target);
            }
            else if (expr is IFieldReferenceExpression)
            {
                processExpression(((IFieldReferenceExpression)expr).Target);
            }
            else if (expr is IObjectCreateExpression)
            {
                processExpression(((IObjectCreateExpression)expr).Initializer);
                foreach (IExpression arg in ((IObjectCreateExpression)expr).Arguments)
                {
                    processExpression(arg);
                }
                MethodDesc callee = getMethodDesc(((IObjectCreateExpression)expr).Constructor);
                if (callee != null)
                {
                    currMethod.addInvocation(
                        callee,
                        false,
                        expr,
                        monitors);
                }
            }
            else if (expr is IBlockExpression)
            {
                foreach (IExpression elem in ((IBlockExpression)expr).Expressions)
                {
                    processExpression(elem);
                }
            }
            else if (expr is IPropertyIndexerExpression)
            {
                processExpression(((IPropertyIndexerExpression)expr).Target);
                foreach (IExpression index in ((IPropertyIndexerExpression)expr).Indices)
                {
                    processExpression(index);
                }
            }
            else if (expr is IPropertyReferenceExpression)
            {
                processExpression(((IPropertyReferenceExpression)expr).Target);
            }
        }


        internal void processMethod(IMethodDeclaration method)
        {
            IMethodDeclaration methodDecl = translator.TranslateMethodDeclaration(method);
            ClassDesc selfClass = getClassDesc(methodDecl.DeclaringType);
            MethodDesc methodDesc = selfClass.getMethod(method);
            MethodDesc saveMethod = currMethod;
            currMethod = methodDesc;

            if (!methodDecl.Abstract)
            {
                MethodInfo mi = methodDecl as MethodInfo;
                if (mi != null)
                {
                    foreach (IMethodReference overrides in methodDecl.Overrides)
                    {
                        IMethodDeclaration overridesDecl = overrides.Resolve();
                        if (overridesDecl != null)
                        {
                            MethodDesc md = getMethodDesc(overridesDecl);
                            if (md != null)
                            {
                                md.addOverriddenBy(methodDesc);
                            }
                        }
                    }
                }
            }

            monitors.clear();
            topLevelLocks.clear();
            IBlockStatement body = methodDecl.Body as IBlockStatement;
            if (body != null)
            {
                processStatement(body);
            }
            methodDesc.topLevelSyncBlocks = topLevelLocks.toArray();
            methodDesc.syncInsns = lockStmts.ToArray();
            if (!topLevelLocks.empty())
            {
                Debug.Assert(methodDesc.nextEntryPoint == null);
                methodDesc.nextEntryPoint = entryPoints;
                entryPoints = methodDesc;
            }
            currMethod = saveMethod;
        }

        internal void processMethod(IMethodReference mth)
        {
            if (mth != null)
            {
                IMethodDeclaration decl = mth.Resolve();
                if (decl != null)
                {
                    processMethod(decl);
                }
            }
        }

        public virtual void message(String msg)
        {
            this.windowManager.ShowMessage(msg);
        }


        internal Dictionary<ITypeDeclaration, ClassDesc> classes;
        internal Dictionary<object, GraphVertex> syncObjects;
        internal List<ILockStatement> lockStmts;
        internal VertexList monitors;
        internal VertexList topLevelLocks;
        internal MethodDesc currMethod;
        internal MethodDesc entryPoints;
        internal bool reportAllLoops;

        static internal CSLintAddin instance;

        void Reset()
        {
            classes = new Dictionary<ITypeDeclaration, ClassDesc>();
            syncObjects = new Dictionary<object, GraphVertex>();
            lockStmts = new List<ILockStatement>();
            monitors = new VertexList();
            topLevelLocks = new VertexList();
            entryPoints = null;
        }

        public CSLintAddin()
        {
            reportAllLoops = true;
            instance = this;
            Reset();
        }
    }
}